import typer
import shutil
import os
from dotenv import load_dotenv
import time

app = typer.Typer()

def starter(project_name) -> bool:
    try:
        from .database.redis_client import redis_client
        print("Loading .env ...")
        print("Loaded .env to system...")
        print("Checking Redis connection...")
        time.sleep(1)
        if redis_client.ping()!=None:
            print("Redis connection established successfully.")
            time.sleep(1)
            # Enable Append Only File
            redis_client.config_set("appendonly", "yes")
            redis_client.config_set("appendfsync", "everysec")
            print("AOF persistence enabled.")
            time.sleep(1)
            return True
        if redis_client.ping()==None:
            print("Redis connection failed.")
            return False
    except Exception as e:
        print("Unable to establish connection with Redis.")
        shutil.rmtree(project_name)
        raise e

@app.command()
def action():
    try:
        project_name = typer.prompt("Project name:->").strip()
        is_redis = typer.prompt("Is Redis currently running? (yes/no): ").strip().lower()
        if is_redis=="yes":
            redis_host = typer.prompt("Enter the Redis host (e.g., 'localhost' for local instance): ").strip().lower()
            redis_port = typer.prompt("Port of Redis:->").strip().lower()
            redis_db_no = typer.prompt("Enter the Redis database number (e.g., '0', '1'): ").strip().lower()
            want_dummy_data = typer.prompt("Do you want to load dummy data for testing? (yes/no): ").strip().lower()
            secret_key = typer.prompt("Do you want to set your own secret key? (Type your unique key(string) or NO): ").strip().lower()
            # Create project folder
            os.makedirs(project_name, exist_ok=True)
            # Create .env file
            env_content = f"""
        REDIS_HOST={redis_host}
        REDIS_PORT={redis_port}
        REDIS_DB_NUMBER={redis_db_no}
                            """
            if secret_key=="yes":
                env_content += f"\nSECRET_KEY={secret_key}"
            elif secret_key=="no":
                print("SECRET_KEY must be set in 'Production' ! Please configure it as an environment variable.")
                time.sleep(2)
            with open(f"{project_name}/.env", "w") as f:
                f.write(env_content)
            # explicitly load the .env file you just created
            env_path = os.path.join(project_name, ".env")
            load_dotenv(dotenv_path=env_path, override=True)
            # Define folders and files
            folders = ["static", "src", "tests", "logs"]
            files = [".gitignore", "requirements.txt", "README.md", "LICENSE"]
            # Create folders
            print(f"Creating Project Structure for: {project_name}")
            time.sleep(1)
            for folder in folders:
                folder_path = os.path.join(project_name, folder)
                os.makedirs(folder_path, exist_ok=True)
            # Create files
            for file in files:
                file_path = os.path.join(project_name, file)
                if not os.path.exists(file_path):
                    with open(file_path, "w") as f:
                        if file == "README.md":
                            f.write("# Project Title\n\nProject description goes here.\n")
                        elif file == "requirements.txt":
                            f.write("# Add project dependencies here\n")
                        elif file == ".gitignore":
                            f.write("__pycache__/\n*.pyc\n.env\n")
                else:
                    print(f"File already exists: {file_path}")
            """
            Check Redis connection
            """
            starter(project_name)
            """
            Creating Required Files & Folders
            """
            if want_dummy_data=="yes":
                from .database.db import Data
                Data.load_data()
                print("Loading default users and roles...")
                time.sleep(1)
            # If everything goes well then your have your Project Folder.
            typer.echo(f"Project {project_name} created with .env file \u2764 \U0001F680")
        elif is_redis=="no":
            print("A running Redis instance is required for Pysible to function.\nPlease start Redis before continuing.")
    except Exception as e:
        return e

def main():
    app()



from .redis_client import redis_client

class Data:

    def load_role():
        try:
            redis_client.hset("role:root", mapping={"name": "root"})
            redis_client.hset("role:admin", mapping={"name": "admin"})
            redis_client.hset("role:editor", mapping={"name": "editor"})
            redis_client.hset("role:viewer", mapping={"name": "viewer"})
        except Exception as e:
            return {f"Failed to load role data: {e}"}
    
    def load_user():
        try:
            redis_client.hset(
            "user_id:root",
            mapping={
                "username": "default_user",
                "password": "unique_password",
                "roles": ",".join(["root", "admin"])
            }
        )
        except Exception as e:
            return {f"Failed to load users: {e}"}

    @staticmethod
    def load_data():
        """
        This will load some initially required data like
        roles - ["root", "admin", "user"]
        and dummy users- ["user1", "user2"]
        *** YOU CAN ADD OR REMOVE THIS DATA ANYTIME YOU WANT ***
        *** YOU CAN ADD NEW ROLES AND USERS ***
        """
        try:
            Data.load_role()
            Data.load_user()
            print("Default roles and admin user initialized.")
        except Exception as e:
            return {f"Failed to execute LOAD DATA function. {e}"}
    
    @staticmethod    
    def create_user(user_id: str, username: str, password: str, roles: list):
        for role in roles:
                if not redis_client.hgetall(f"role:{role}"):
                    print(f"Role '{role}' not found. Please create the role first using create_role().")
                    return False
        if redis_client.keys(f"user_id:{user_id}"):
            print("User ID already exists. Please choose a different identifier.")
            return False
        try:
            from ..logger import logger
            redis_client.hset(
                f"user_id:{user_id}",
                mapping={
                    "username": username,
                    "password": password,
                    "roles": ",".join(roles) 
                }
            )
            logger.info(f"New user created. User ID: {user_id}, Username: {username}")
            print("User successfully saved to the database.")
            return True
        except Exception as e:
            return {f"Failed to add new user {e}"}
    
    @staticmethod
    def create_role(role: str):
        if redis_client.keys(f"role:{role}"):
            print("Role already exists in the database. Skipping creation.")
            return False
        try:
            from ..logger import logger
            redis_client.hset(f"role:{role}", mapping={"name": f"{role}"})
            logger.info(f"New role created. Role: {role}")
            print("Role successfully saved to the database.")
            return True
        except Exception as e:
            raise e



import time
from ..database.redis_client import redis_client
from fastapi import Request, HTTPException, Depends
from .rbac import RBAC
from ..logger import logger

class PyRate:

    @staticmethod
    def rate_limiter(rate: int, burst: int):
        """
        Returns a dependency function for FastAPI
        """
        def dependency(request: Request):
            client_ip = request.client.host
            key_tokens = f"tokens:{client_ip}"
            key_timestamp = f"timestamp:{client_ip}"
            now = time.time()

            tokens = redis_client.get(key_tokens)
            tokens = float(tokens) if tokens else burst

            last_refill = redis_client.get(key_timestamp)
            last_refill = float(last_refill) if last_refill else now

            elapsed = now - last_refill
            tokens = min(burst, tokens + elapsed * rate)

            if tokens < 1:
                logger.warning(f"'Too many requests' to endpoint: '{request.url.path}'")
                raise HTTPException(status_code=429, detail="Too many requests")

            tokens -= 1
            redis_client.set(key_tokens, tokens)
            redis_client.set(key_timestamp, now)

        return dependency





from ..database.redis_client import redis_client
from fastapi.responses import JSONResponse
from .token import Token
from fastapi import HTTPException, status
from ..logger import logger

class Auth:

    @staticmethod
    def login(form_data):
        user_key = f"user_id:{form_data.username}"
        # Check if user exists
        if not redis_client.exists(user_key):
            logger.warning(f"Login failed | Username: {form_data.username} | Reason: user not found")
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Invalid username or password."
            )
        stored_password = redis_client.hget(user_key, "password")
        if not stored_password:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Invalid username or password."
            )
        access_token = Token.create_token(user_id=form_data.username)
        response = JSONResponse(content={"message": "Login successful."})
        response.set_cookie(
            key="token",
            value=access_token,
            httponly=True,
            secure=True,      # important for production
            samesite="Strict" # prevents CSRF
        )
        logger.info(f"User '{form_data.username}' logged in successfully.")
        return response
        
    @staticmethod
    def logout(user_id: str ):
        if not user_id:
            logger.warning("Logout attempt without authentication.")
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Authentication required. Please log in first."
            )
        response = JSONResponse(content={"message": f"User '{user_id}' logged out successfully."})
        response.delete_cookie(key="token")
        logger.info(f"User '{user_id}' logged out successfully.")
        return response
            


import os
from dotenv import load_dotenv
import redis

load_dotenv()

REDIS_HOST = os.getenv("REDIS_HOST")
REDIS_PORT = os.getenv("REDIS_PORT")
REDIS_DB_TO_USE = os.getenv("REDIS_DB_NUMBER")

redis_client = redis.Redis(host=REDIS_HOST, port=REDIS_PORT, db=REDIS_DB_TO_USE)




from fastapi import Depends, HTTPException, status, Request
from ..database.redis_client import redis_client
from .token import Token
from ..logger import logger

class RBAC:
    """
    class methods to deal with RBAC.
    Recieve "roles" (list) as a parameter and match with redis db.
    """
    @staticmethod
    def require_token(request: Request):
        token = request.cookies.get("token")
        if not token:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Authentication token is missing."
            )
        payload = Token.decode_token(token=token)  # your custom decode wrapper
        user_id: str = payload.get("user_id")
        if not user_id:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Token is missing required user information."
            )
        return user_id
    
    @staticmethod
    def require_role(roles: list):
        def role_checker(
            cred_user_id: str = Depends(RBAC.require_token),
            request: Request = None
        ):
            stored_roles = redis_client.hget(f"user_id:{cred_user_id}", "roles")
            if not stored_roles:
                logger.warning(f"User '{cred_user_id}' has no roles assigned | Endpoint: '{request.url.path}'")
                raise HTTPException(
                    status_code=status.HTTP_403_FORBIDDEN,
                    detail="No roles assigned. Access denied."
                )
            user_roles = stored_roles.decode().split(",")
            if any(role in user_roles for role in roles):
                return cred_user_id
            logger.warning(
                f"'Unauthorized' access attempt | User: '{cred_user_id}' | Endpoint: '{request.url.path}'"
            )
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail="Insufficient permissions to access this endpoint."
            )
        return role_checker


from datetime import datetime, timedelta
from jose import jwt, JWTError
import os

DEFAULT_SECRET_KEY = ".P-L-E-A-S-E-@-C-H-A-N-G-E-@-T-H-I-S-@-K-E-Y."
SECRET_KEY = os.getenv("SECRET_KEY", DEFAULT_SECRET_KEY)
ALGORITHM = "HS256"
ACCESS_EXPIRE_MIN = 30

class Token:
    
    @staticmethod
    def create_token(user_id: str):
        expire = datetime.now() + timedelta(minutes=ACCESS_EXPIRE_MIN)
        to_encode = {"user_id": user_id, "exp": expire}
        return jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)
    
    @staticmethod
    def decode_token(token: str):
        try:
            return jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
        except JWTError:
            return None